//---------------------------------------------------------------------------
// Copyright 1997 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License (GPL).
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// Additionally this source is licensed exclusively to the company
// Bruker Analytik GmbH under the terms of the GNU Lesser General Public 
// License (LGPL). See the file 'COPYING.LIB' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/STL/String.h,v $
// $Id: String.h,v 1.4 2005/09/25 23:44:37 ah Exp $
//
// Description: utility functions for working with string(STL) objects.
//
//---------------------------------------------------------------------------

#ifndef AH_STL_STRING_H		// prevent multiple includes
#define AH_STL_STRING_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
USING_NAMESPACE(AH)

#include <string>
#include <vector>
USING_NAMESPACE(std)

// --- types and classes ---

BEGIN_NAMESPACE(AH)

#define AH_WHITE_SPACE " \t\n\r"

/// extract all tokens from a STL string in one call
/**
 {\tt STLStringToken()} works on a {\tt string}(STL) like {\tt strtok(3S)} 
 on C type character arrays. Other than {\tt strtok()} all the work is done 
 in one call, storing the tokens in the string vector  {\tt result}.

 {\tt UtilStringToken()} does not clear the string vector, so this must be 
 done before the function call.
 */ 
void STLStringToken
(
    /// string to extract tokens from
    const string&	line,		
    /// sequential string container
    vector<string>&	result,
    /// token separator characters (default is white space)
    const char*		separator = AH_WHITE_SPACE
);



/// removes leading characters from {\tt line}
/**
 {\tt STLStringDelLead()} removes all leading {\tt characters} in 
 {\tt line}. The set of characters to be removed can be specified.
 Without explicit specification white space is removed.
 */ 
extern void STLStringDelLead
(
    /// String where the characters are removed
    string&	line,
    /// which characters to remove (default is white space)
    const char*	remove = AH_WHITE_SPACE
);


/// removes following characters from {\tt line}
/**
 {\tt STLStringDelFollow()} removes all following characters in 
 {\tt line}. The set of characters to be removed can be specified.
 Without explicit specification white space is removed.
 */ 
extern void STLStringDelFollow
(
    /// String where the characters are removed
    string&	line,
    /// which characters to remove (default is white space)
    const char*	remove = AH_WHITE_SPACE
);


/// removes leading and following characters from {\tt line}
/**
 {\tt STLStringDelSurround()} removes all leading and following 
 characters in {\tt line}. The set of characters to be removed can be 
 specified. Without explicit specification white space is removed.
 */ 
extern  void STLStringDelSurround
(
    /// String where the characters are removed
    string&	line,
    /// which characters to remove (default is white space)
    const char*	remove = AH_WHITE_SPACE
);


/** convert printf(3S) compatible format string into STL string.

 {\tt STLStringPrintf()} converts a printf(3S) compatible format string
 and all arguments into a STL string. The maximum size is 1024 characters
 including the terminating 0. If this size is too short then use the
 other version of {\tt STLStringPrintf()} which accepts the desired size 
 as first argument.

 If the size of the buffer is exceeded the returned string is either 
 empty or truncated (depends on the C99 compatibility of the C++ compiler).

 @param maxSize	maximum size of message (number of characters)
 @param format	printf(3S) compatible format string
 @param ...	remaining arguments according to format string
 */
string STLStringPrintf
(
    unsigned int maxSize,
    const char* format,
    ...
);


// --- inline function implementation ---


inline void STLStringDelSurround
(
    string&	line,
    const char*	remove
)
{
    STLStringDelLead(line, remove);
    STLStringDelFollow(line, remove);
}


END_NAMESPACE
#endif // !AH_STL_STRING_H



