//---------------------------------------------------------------------------
// Copyright 1999 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHio/RiffWaveFile.c,v $
// $Id: 
//
// Description: 
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: RiffWaveFile.c,v 1.29 2005/10/31 01:20:07 ah Exp $ (c) 1999 Andreas Huennebeck";

// --- necessary includes ---

#include <AH/IO/Riff/Chunk.h>
#include <AH/IO/Riff/StartChunk.h>
#include <AH/IO/Riff/WaveFmt.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/Error/Intern.h>
#include <AH/Error/Sys.h>
#include <AH/Error/Basic.h>
#include <AH/IO/Smart/Buffer.h>
#include <AH/IO/Smart/FilePtr.h>
#include <AH/IO/CFormat.h>
#include <AH/IO/Convert.h>
USING_NAMESPACE(AH)

#include <cstdio>
#include <string>
#include <iostream>
USING_NAMESPACE(std)


BEGIN_NAMESPACE(AH)

typedef IOCFormat Cfm;

    //-------------------
    // con-/destructors:
    //-------------------
    

// constructor for input WAV file 
IORiffWaveFile::IORiffWaveFile(const char* name, bool isPipe, int v)
    : waveName(0), verbose(v), waveFp(0), waveIsPipe(isPipe), pipePos(0)
{
    readHeader(name, isPipe);
}



// constructor for input file (WAV, MP3, Ogg, ...) 
IORiffWaveFile::IORiffWaveFile(const char* name, int v)
    : waveName(0), verbose(v), waveFp(0), waveIsPipe(false), pipePos(0)
{
    // determine basename and extension of input file:
    std::string inputFileName(name);
    std::string extension;
    std::string baseName;
    string::size_type pos = inputFileName.find_last_of(".");
    if (string::npos != pos) 
    {
	baseName = inputFileName.substr(0, pos);
	extension = inputFileName.substr(pos+1);
    }
    std::string dirName;
    pos = baseName.find_last_of("/\\");
    if (string::npos != pos) 
    {
	dirName = baseName.substr(0, pos);
	baseName = baseName.substr(pos+1);
    }

    // handle each file according to its extension:
    std::string inCmd;
    std::string devnull(" 2> /dev/null");
    if (0 == strcasecmp(extension.c_str(), "mp3"))
    {
	// use mpg123 to convert MP3 file in WAV
	inCmd = "mpg123 -q -w - \"" + inputFileName + "\"" + devnull;
	waveIsPipe = true;
	if (verbose) cerr << "reading MP3 file '" << inputFileName
			  << "'" << endl;
    }
    else if (0 == strcasecmp(extension.c_str(), "ogg"))
    {
	// use oggdec to convert ogg file in WAV
        inCmd = "oggdec -Q -o - \"" + inputFileName + "\"" + devnull;
	waveIsPipe = true;
	if (verbose) cerr << "reading OGG-Vorbis file '" << inputFileName
			  << "'" << endl;
    }
    else if ((0 == strcasecmp(extension.c_str(), "mpc")) ||
	     (0 == strcasecmp(extension.c_str(), "mp+")) ||
	     (0 == strcasecmp(extension.c_str(), "mpp")))
    {
	// use mppdec to convert Mpeg+ file in WAV
        inCmd = "mppdec --silent - - < \"" + inputFileName + "\"" + devnull;
	waveIsPipe = true;
	if (verbose) cerr << "reading MPEG+ file '" << inputFileName
			  << "'" << endl;
    }
    else if (0 == strcasecmp(extension.c_str(), "pac"))
    {
	// use lpac to convert pac file in WAV
        inCmd = "lpac -x \"" + inputFileName + "\" - " + devnull;
	waveIsPipe = true;
	if (verbose) cerr << "reading PAC file '" << inputFileName
			  << "'" << endl;
    }
    else
    {
	inCmd = inputFileName; // default is a WAV file
	if (verbose) cerr << "reading WAV file '" << inputFileName
			  << "'" << endl;
    }
    if (waveIsPipe && (verbose > 1))
	cerr << "call '" << inCmd << "'" << endl;

    // create a wav file object for read:
    if (3 < verbose) cerr << "open input file \"" << inputFileName
			  << "\":" << endl;
    readHeader(inCmd.c_str(), waveIsPipe);
}



// constructor for output WAV file 
IORiffWaveFile::IORiffWaveFile
(
    const char* name,	// name of output file
    uint16 channel_num,	// number of channels
    uint16 sample_size,	// size of sample in bits
    uint32 sample_rate,	// sample rate in Hz
    uint32 sample_num,	// number of samples per channel
    int v
)
    : waveName(0), verbose(v), waveFp(0), waveIsPipe(false),
      waveSamples(sample_num), channels(channel_num),
      sampleRate(sample_rate), sampleSize(sample_size),
      maxPosAmpl((1 << (sampleSize - 1)) - 1), maxNegAmpl(-maxPosAmpl -1)
{
    writeHeader(name);
}



// destructor
IORiffWaveFile::~IORiffWaveFile()
{
    // close open file:
    if (waveFp && (waveFp != stdout))
	(void)fclose(waveFp);
}


    //-----------------------
    // read/write functions:
    //-----------------------


// read header of wave file
void IORiffWaveFile::readHeader(const char* name, bool isPipe)
{
    if (!name) throw ErrorInternNullPtrArg();

    // close open file:
    if (waveFp) (void)fclose(waveFp);

    //--------------------------------
    // open new file for binary read:
    //--------------------------------
    waveName = name;
    waveIsPipe = isPipe;
    if (waveIsPipe)
	waveFp = popen(waveName, "r");
    else if ((0 == strcmp(waveName, "stdin")) ||
	     (0 == strcmp(waveName, "-")) ||
	     (0 == strcmp(waveName, "stdin")))
	waveFp = stdin;
    else
	waveFp = fopen(waveName, "r");
    if (!waveFp) throw ErrorSys(waveName);
    fileStart = getFilePos();

    //-------------------
    // read start chunk:
    //-------------------
    IOSmartBuffer<unsigned char> buf(IORiffStartChunk::size);
    readBuf(buf(), buf.getSize());
    IORiffStartChunk startChunk(buf());
    waveFileSize = startChunk.getChunkSize();
    if (waveIsPipe || (pipePos == waveFileSize))
	waveFileSize = 0x7fffffff - pipePos;
    fileEnd = fileStart + waveFileSize;

    //---------------------------
    // read format chunk header:
    //---------------------------
    buf.resize(IORiffChunk::size);
    readBuf(buf(), buf.getSize());
    IORiffChunk formatchunk(buf());

    //-------------------------
    // read format chunk data:
    //-------------------------
    buf.resize(formatchunk.getDataSize());
    readBuf(buf(), buf.getSize());
    IORiffWaveFmt fmt(buf());
    channels = fmt.getChannels();
    sampleRate = fmt.getSampleRate();
    sampleSize = fmt.getSampleSize();
    maxPosAmpl = (1 << (sampleSize - 1)) - 1;
    maxNegAmpl = -maxPosAmpl -1;

    //-------------------------
    // read data chunk header:
    //-------------------------
    buf.resize(IORiffChunk::size);
    readBuf(buf(), buf.getSize());
    IORiffChunk datachunk(buf());

    waveDataSize = datachunk.getDataSize(); 
    if (waveIsPipe || (0 == waveDataSize))
	waveDataSize = 0x7fffffff - pipePos;
    if (sampleSize) 
	waveSamples = waveDataSize / (channels * sampleSize/8);
    else
	waveSamples = 0; // this should not happen!

    waveTime = getTime(waveSamples);
    dataStart = getFilePos();
    dataEnd = dataStart + waveDataSize;
    curDataStart = dataStart;
    curDataEnd = dataEnd;
}


// read PCM samples from wave file
int32 IORiffWaveFile::readSamples(int32* data, uint32 num)
{
    const uint32 bufSize = getDataBufSize(num);
    unsigned char buf[bufSize];

    // read data into buffer:
    int32 rdsize = readRaw(buf, bufSize);
    if (0 >= rdsize) return rdsize;

    // get samples from buffer:
    if ((uint32)rdsize < bufSize)
	num = rdsize / getDataBufSize((uint32)1);
    for (uint32 i = 0; i < num; ++i)
    {
	for (uint32 chan = 0; chan < getChannelNum(); ++chan)
	{
	    int32 value = getBufSample(buf, i, chan+1); 
	    *data++ = value;
	}
    }

    return num;
}


// read raw PCM data from wave file
int32 IORiffWaveFile::readRaw(unsigned char* data, uint32 size)
{
    int32 ret = readBuf(data, size);
    curDataStart = getFilePos();
    return ret;
}



// read a data buffer from wave file
int32 IORiffWaveFile::readBuf(unsigned char* buf, uint32 size)
{
    errno = 0;
    int32 ret = fread(buf, 1, size, waveFp);
    if (0 >= ret && !feof(waveFp))
    {
	if (errno) throw ErrorSys(waveName);
	throw ErrorBasic(string("cannot read from ") + waveName);
    }
    pipePos += ret;
    return ret;
}



// write header into new wave file
void IORiffWaveFile::writeHeader(const char* name)
{
    if (!name) throw ErrorInternNullPtrArg();

    // close open file:
    if (waveFp && (waveFp != stdout))
	(void)fclose(waveFp);

    waveDataSize = waveSamples * (channels * sampleSize/8);
    waveTime = getTime(waveSamples);

    //--------------------------------
    // open new file for binary read:
    //--------------------------------
    waveName = name;
    if (0 == strcmp("-", waveName)) 
	waveFp = stdout;
    else
    {
	waveFp = fopen(waveName, "w");
	if (!waveFp) throw ErrorSys(waveName);
    }
    if (waveFp == stdout) fileStart = 0;
    else fileStart = getFilePos();

    //---------------------------
    // create header:
    // - create data chunk header:
    //---------------------------
    IORiffChunk datachunk("data", 0, waveDataSize);
    waveFileSize = datachunk.getChunkSize();

    //-----------------------------
    // - create format chunk data:
    //-----------------------------
    IORiffWaveFmt fmtdata(IORiffWaveFmt::MS_PCM, 
			  channels, sampleRate, sampleSize);
    waveFileSize += fmtdata.getDataSize();

    //-------------------------------
    // - create format chunk header:
    //-------------------------------
    IORiffChunk fmtchunk("fmt ", 0, fmtdata.getDataSize());
    waveFileSize += fmtchunk.getChunkSize();

    //------------------------
    // - create start chunk :
    //------------------------
    IORiffStartChunk startchunk(waveFileSize, "WAVE");
    waveFileSize += startchunk.getChunkSize();

    //--------------------------------
    // write header: 
    // - write start chunk
    // - write format chunk header
    // - write format chunk data
    // - write data chunk header
    //--------------------------------
    writeBuf(startchunk.chunkBuffer(), IORiffStartChunk::size);
    writeBuf(fmtchunk.chunkBuffer(), IORiffChunk::size);
    writeBuf(fmtdata.getDataBuffer(), fmtdata.getDataSize());
    writeBuf(datachunk.chunkBuffer(), IORiffChunk::size);

    // keep track of begin and end of data: 
    if (waveFp == stdout) dataStart = waveFileSize;
    else dataStart = getFilePos();
    dataEnd = dataStart + waveDataSize;
    curDataStart = dataStart;
    curDataEnd = dataEnd;
}



// rewrite header into current wave file
void IORiffWaveFile::writeHeader()
{
    if (waveFp == stdout) return;

    // get current file position (assume at end of file)
    waveFileSize = getDataPos();
    waveDataSize = waveFileSize - dataStart;
    waveSamples = waveDataSize / (channels * sampleSize/8);
    waveTime = getTime(waveSamples);

    // set file pointer to begin of file: 
    fileStart = setFilePos((uint32)0);

    //-------------------------------
    // create header:
    // - create data chunk header
    // - create format chunk data
    // - create format chunk header
    // - create start chunk 
    //-------------------------------
    IORiffChunk datachunk("data", 0, waveDataSize);
    IORiffWaveFmt fmtdata(IORiffWaveFmt::MS_PCM, 
			  channels, sampleRate, sampleSize);
    IORiffChunk fmtchunk("fmt ", 0, fmtdata.getDataSize());
    IORiffStartChunk startchunk(waveFileSize, "WAVE");

    //--------------------------------
    // write header: 
    // - write start chunk
    // - write format chunk header
    // - write format chunk data
    // - write data chunk header
    //--------------------------------
    writeBuf(startchunk.chunkBuffer(), IORiffStartChunk::size);
    writeBuf(fmtchunk.chunkBuffer(), IORiffChunk::size);
    writeBuf(fmtdata.getDataBuffer(), fmtdata.getDataSize());
    writeBuf(datachunk.chunkBuffer(), IORiffChunk::size);

    // keep track of begin and end of data: 
    dataStart = getFilePos();
    dataEnd = dataStart + waveDataSize;
    curDataStart = dataStart;
    curDataEnd = dataEnd;
}



// write PCM samples into wave file
int32 IORiffWaveFile::writeSamples(const int32* data, uint32 num)
{
    const uint32 bufSize = getDataBufSize(num);
    unsigned char buf[bufSize];

    // put samples into buffer
    for (uint32 i = 0; i < num; ++i)
    {
	for (uint32 chan = 0; chan < getChannelNum(); ++chan)
	{
	    int32 value = *data++;
	    setBufSample(buf, i, chan+1, value); 
	}
    }

    // write data :
    int32 wrsize = writeRaw(buf, bufSize);
    if (0 >= wrsize) return wrsize;
    if ((uint32)wrsize < bufSize)
	num = wrsize / getDataBufSize((uint32)1);

    return num;
}


// write raw PCM data into new wave file
int32 IORiffWaveFile::writeRaw(const unsigned char* data, uint32 size)
{
    //---------------------------
    // write raw data:
    //---------------------------
    int32 ret = writeBuf(data, size); 
    if (waveFp != stdout) curDataStart = getFilePos();
    return ret;
}


// write a data buffer into wave file
int32 IORiffWaveFile::writeBuf(const unsigned char* buf, uint32 size)
{
    //---------------------------
    // write raw data:
    //---------------------------
    errno = 0;
    int32 ret = fwrite(buf, 1, size, waveFp); 
    if (0 > ret)
    {
	if (errno) throw ErrorSys(waveName);
	throw ErrorBasic(string("cannot write to ") + waveName);
    } 
    return ret;
}


    //------------------
    // file positioning
    //------------------


// get current data position
uint32 IORiffWaveFile::getDataPos() const
{
    return getFilePos();
}



// get data position from time/sec
uint32 IORiffWaveFile::getDataPos(double t) const
{
    uint32 filePos = uint32(t * 
			    (double)(channels * sampleSize/8 * sampleRate));
    return (dataStart + filePos);
}



// get data position from specified sample number
uint32 IORiffWaveFile::getDataPos(uint32 sampleNum) const
{
    uint32 filePos = sampleNum * (channels * sampleSize/8);
    return (dataStart + filePos);
}



// set file to sample position specified by time/sec
uint32 IORiffWaveFile::setDataPos(double t)
{
    return setFilePos(getDataPos(t));
}



// set file to specified sample number
uint32 IORiffWaveFile::setDataPos(uint32 sampleNum)
{
    uint32 pos = (channels * sampleSize/8) * sampleNum;
    pos += dataStart;
    return setFilePos(pos);
}



// get current file position
uint32 IORiffWaveFile::getFilePos() const
{
    if (waveIsPipe) return pipePos;

    long pos = ftell(waveFp);
    if (-1 == pos) throw ErrorSys(waveName);
    return (uint32)pos;
}



// set file position
uint32 IORiffWaveFile::setFilePos(uint32 pos)
{
    if (waveIsPipe)
    {
	// seek is not possible with pipes:
	if (pos == pipePos) return pos;
	throw ErrorBasic(std::string(waveName) +
			 ": cannot set file position in pipe");
    }
    if (-1 == fseek(waveFp, (long)pos, SEEK_SET))
	throw ErrorSys(waveName);
    return pos;
}


    //----------------------
    // conversion utilities
    //----------------------


// get number of samples from time/sec
uint32 IORiffWaveFile::getSampleNum(double t) const
{
    return (uint32)(t * sampleRate);
}



// get time/sec from number of samples
double IORiffWaveFile::getTime(uint32 sampleNum) const
{
    return ((double)sampleNum / (double)sampleRate);
}


    //-------------------
    // buffer utilities:
    //-------------------


// get size of data for desired time/sec
uint32 IORiffWaveFile::getDataBufSize(double t) const
{
    return getDataBufSize(getSampleNum(t));
}



// get size of data for desired samples
uint32 IORiffWaveFile::getDataBufSize(uint32 sampleNum) const
{
    return ((channels * sampleSize/8) * sampleNum);
}



// get specified sample from buffer
int32 IORiffWaveFile::getBufSample
(
    unsigned char* buffer,
    uint32 sampleNum, 
    int channel
)
{
    uint32 offset = getDataBufSize(sampleNum);
    uint8* ptr = buffer + offset;               // begin of samples
    ptr += sampleSize/8 * (channel - 1);        // begin of our sample

    // samples are stored in little endian:
    if (sampleSize == 16) return (int32)IOConvert::LittleEndCharsToInt16(ptr);
    if (sampleSize == 8)  return (int32)(*ptr);
    if (sampleSize == 32) return IOConvert::LittleEndCharsToInt32(ptr);

    // we use fast multiplication: (X * 8) == (X << 3)
    uint32 val = 0;
    for (int i = 0; i < sampleSize/8; ++i)
	val += (*ptr++) << (i << 3);
    if (val & (1 << (sampleSize-1)))
	val |= 0xffffffff << sampleSize-1;    // handle sign

    return (int32)val;
}

// set specified sample in buffer
void IORiffWaveFile::setBufSample
(
    unsigned char* buffer,
    uint32 sampleNum,
    int channel,
    int32 sampleValue
)
{
    // check limits:
    if (sampleValue > maxPosAmpl)      sampleValue = maxPosAmpl;
    else if (sampleValue < maxNegAmpl) sampleValue = maxNegAmpl;

    uint32 offset = getDataBufSize(sampleNum);
    uint8* ptr = buffer + offset;               // begin of samples
    ptr += sampleSize/8 * (channel - 1);        // begin of our sample

    // samples are stored in little endian:
    if (sampleSize == 16)
	IOConvert::Uint16ToLittleEndChars(sampleValue, ptr);
    else if (sampleSize == 8)
	*((int8*)ptr) = sampleValue;
    else if (sampleSize == 32)
	IOConvert::Uint32ToLittleEndChars(sampleValue, ptr);
    else 
    {
	// to be checked - see function above (is bad):
	// we use fast multiplication: (X * 8) == (X << 3)
	for (int i = 0; i < sampleSize/8; i++)
	    *ptr++ = (sampleValue >> ((i << 3))) & 0xff;
    }
}


    //---------------------
    // analyzer functions:
    //---------------------


// analyze time frame
void IORiffWaveFile::getLimits
(
    /// starting point
    double t1,
    /// end point
    double t2,
    int16&	lMin,
    int16&	lMax,
    int16&	rMin,
    int16&	rMax,
    int32&	lMinCnt,
    int32&	lMaxCnt,
    int32&	rMinCnt,
    int32&	rMaxCnt
)
{
    const uint32 samples_per_block = 65536;

    //------------------
    // check arguments:
    //------------------
    if (t1 > t2)
    {
	double tmp = t1;
	t1 = t2;
	t2 = tmp;
    }
    if (t1 < 0) t1 = 0;
    else if (t1 >= waveTime) throw ErrorInternBadArg();
    if (t2 > waveTime) t2 = waveTime;
    else if (t2 <= 0) throw ErrorInternBadArg();

    //-------------------------------------------------------
    // prepare maximum/minimum values for whole time frame:
    //-------------------------------------------------------
    lMax = -32766;
    rMax = -32766;
    lMin = 32767;
    rMin = 32767;

    //------------------------------------------
    // prepare maximum/minimum value counters:
    //------------------------------------------
    lMaxCnt = 0;
    rMaxCnt = 0;
    lMinCnt = 0;
    rMinCnt = 0;

    //-----------------------------
    // set file to start position:
    //-----------------------------
    setFilePos(getDataPos(t1));

    //---------------------------------
    // get number of samples to read:
    //---------------------------------
    size_t samples_to_read = getSampleNum(t2 - t1);

    IOSmartBuffer<int16> buf;
    do
    {
	uint32 sample_num;
	if (samples_to_read > samples_per_block)
	    sample_num = samples_per_block;
	else
	    sample_num = samples_to_read;
	buf.resize(sample_num * channels);

	//----------------------------------------
	// read whole memory in steps:
	//----------------------------------------
	readBuf((unsigned char*)buf(), buf.getSize());

	//----------------------------------------
	// get min/max values from current block:
	//----------------------------------------
	for (size_t cnt = 0; cnt < sample_num; cnt++)
	{
	    int16 lval = buf.getEntry(2*cnt); 
	    int16 rval = buf.getEntry(2*cnt+1); 
	    if (lMax < lval) 
	    {
		lMax = lval;	// catch new maximum
		lMaxCnt = 1;	// reset counter
	    }
	    else if (lMax == lval) lMaxCnt++;

	    if (lMin > lval)
	    {
		lMin = lval;	// catch new minimum
		lMinCnt = 1;	// reset counter
	    }
	    else if (lMin == lval) lMinCnt++;

	    if (rMax < rval) 
	    {
		rMax = rval;	// catch new maximum
		rMaxCnt = 1;	// reset counter
	    }
	    else if (rMax == rval) rMaxCnt++;

	    if (rMin > rval)
	    {
		rMin = rval;	// catch new minimum
		rMinCnt = 1;	// reset counter
	    }
	    else if (rMin == rval) rMinCnt++;
	}	
	samples_to_read -= sample_num; 
    } while (samples_to_read);
}


    //-----------------
    // info functions:
    //-----------------

// print information about the wave file
ostream& IORiffWaveFile::printInfo(ostream& out)
{
    out << waveName << ":" << endl
	<< "  " << waveFileSize << " bytes, data = " << waveDataSize
	<< " bytes" << endl 
	<< "  " << waveSamples << " samples = " 
	<< IOConvert::timeToString(waveTime) << " ("
	<< IOConvert::timeToCdString(waveTime) << ")" << endl
        << "  " << channels << " channel" << (channels > 1 ? "s":"") 
	<< " at " << sampleRate << " samples/sec" 
	<< " with " << sampleSize << "bits/sample"
	<< endl;
    return out;
}

// print data for specific time frame
ostream& IORiffWaveFile::printData
(
    ostream& out, 
    size_t ptnum,
    double t1,
    double t2
)
{
    //------------------
    // check arguments:
    //------------------
    if (t1 > t2)
    {
	double tmp = t1;
	t1 = t2;
	t2 = tmp;
    }
    if (t1 < 0) t1 = 0;
    else if (t1 >= waveTime) throw ErrorInternBadArg();
    if (t2 > waveTime) t2 = waveTime;
    else if (t2 <= 0) throw ErrorInternBadArg();

    //-----------------------------
    // set file to start position:
    //-----------------------------
    setFilePos(getDataPos(t1));

    //-----------------------------
    // get size of memory to read:
    //-----------------------------
    size_t samples_to_read = getSampleNum(t2 - t1);
    uint32 samples_per_block = samples_to_read / ptnum;

    Cfm posT("%10.4f");
    Cfm posP("%10d");
    Cfm dat("%6d");
    if (samples_per_block < 2)
    {
	//--------------------------------
	// read whole memory in one step:
	//--------------------------------
	IOSmartBuffer<int16> buf(samples_to_read * channels);
	readBuf((unsigned char*)buf(), buf.getSize());
	for (size_t cnt = 0; cnt < samples_to_read; cnt++)
	{
	    int16 lval = buf.getEntry(2*cnt); 
	    int16 rval = buf.getEntry(2*cnt+1); 
	    out << posP << cnt << posT << t1 + (double)cnt/(double)sampleRate
		<< " " << dat << lval 
		<< " " << dat << rval 
		<< " " << dat << lval - rval 
		<< endl;
	}	
    }
    else
    {
	//----------------------------------------
	// read whole memory in 'ptnum' steps:
	//----------------------------------------
	IOSmartBuffer<int16> buf(samples_per_block * channels);
	for (size_t block = 0; block < ptnum; block++)
	{
	    buf.clear();
	    readBuf((unsigned char*)buf(), buf.getSize());

	    //----------------------------------------
	    // get min/max values from current block:
	    //----------------------------------------
	    int16 lmax = -32766;
	    int16 rmax = -32766;
	    int16 lmin = 32767;
	    int16 rmin = 32767;
	    for (size_t cnt = 0; cnt < samples_per_block; cnt++)
	    {
		int16 lval = buf.getEntry(2*cnt); 
		int16 rval = buf.getEntry(2*cnt+1); 
		if (lmax < lval) lmax = lval;
		if (lmin > lval) lmin = lval;
		if (rmax < rval) rmax = rval;
		if (rmin > rval) rmin = rval;
	    }	
	    //------------------------------------------
	    // print min/max values from current block:
	    //------------------------------------------
	    unsigned int sample_no = block * samples_per_block;
	    double tcur = t1 + getTime(sample_no);
	    out << posP << sample_no << posT << tcur 
		<< " " << dat << lmax 
		<< " " << dat << rmax
		<< " " << dat << lmax - rmax 
		<< endl;
	    out << posP << sample_no << posT << tcur 
		<< " " << dat << lmin 
		<< " " << dat << rmin 
		<< " " << dat << lmin - rmin 
		<< endl;
	}
    }
    return out;
}




END_NAMESPACE
