#include <AH/IO/AudioFile.h>
#include <AH/CUtil/SecureArray.h>
#include <AH/Test/UnitTest.h>

using namespace AH;

//-----------------
// our test class:
//-----------------
class Test: public UnitTest
{
    // basic test functions, to be called in higher level test functions
    void test_simple_file(const std::string& filename, int samples, int channels, int rate);
    void test_converted_file(const std::string& filename, int channels, int rate);

    // test functions, to be called in Test::run():
    void test_all_simple_files();
    void test_all_converted_files();
    void test_missing_file();

    // run all tests:
    virtual void run(const std::vector<std::string>&)
    {
	this->test_all_simple_files();
	this->test_all_converted_files();
	this->test_missing_file();
    }

 public:
    // constructor
    Test(): UnitTest("class IOAudioFile", std::cout) {}

}; 

UnitTest* theUnitTest =  new Test();


void Test::test_simple_file(const std::string& filename, int samples, int channels, int rate)
{ 
    this->announceTestSeries("file " + filename);

    // run the tests:
    UNIT_TEST_BEGIN

    this->announceNextTest("constructor");
    IOAudioFile<int32> af(filename, this->isVerbose());
    this->printPassed();

    this->checkEqual("getNumSamples()", af.getNumSamples(), samples);
    this->checkEqual("getNumChannels()", af.getNumChannels(), channels);
    this->checkEqual("getSampleRate()", af.getSampleRate(), rate);

    this->announceNextTest("getInfo()");
    std::string info = af.getInfo();
    this->printPassed();

    this->checkTrue("getInfo() != 0", info != "");
    if (this->isVerbose()) this->out << info;

    UNIT_TEST_END
} // end of Test::test_simple_file()



void Test::test_all_simple_files()
{ 
    this->test_simple_file("test_01.wav", 10, 1, 96000);
    this->test_simple_file("test_02.wav", 20, 2, 48000);
} // end of Test::test_all_simple_files()



void Test::test_converted_file(const std::string& filename, int channels, int rate)
{ 
    this->announceTestSeries("file " + filename);

    // run the tests:
    UNIT_TEST_BEGIN

    this->announceNextTest("constructor IOAudioFileOpener");
    IOAudioFileOpener afo;
    this->printPassed();

    this->announceNextTest("constructor IOAudioFile");
    IOAudioFile<int32> af(filename, afo, this->isVerbose());
    this->printPassed();

    this->checkEqual("getNumChannels()", af.getNumChannels(), channels);
    this->checkEqual("getSampleRate()", af.getSampleRate(), rate);

    this->announceNextTest("getInfo()");
    std::string info = af.getInfo();
    this->printPassed();

    this->checkTrue("getInfo() != 0", info != "");
    if (this->isVerbose()) this->out << info;

    UNIT_TEST_END
} // end of Test::test_converted_file()



void Test::test_all_converted_files()
{ 
    this->test_converted_file("test_03.mp3", 1, 44100);
} // end of Test::test_all_converted_files()



void Test::test_missing_file()
{ 
    std::string filename = "nonexisting.wav";
    this->announceTestSeries("missing file " + filename);

    // run the test:
    UNIT_TEST_EXCEPTION_BEGIN(ErrorSys)

    IOAudioFile<int32> af(filename, this->isVerbose());

    UNIT_TEST_EXCEPTION_END(ErrorSys)
} // end of Test::test_missing_file()



