//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/wavtool/pcm2wav.c,v $
// $Id: pcm2wav.c,v 1.2 2004/04/20 22:43:37 ah Exp $ 
//
// Description: convert endless PCM data into one or several wav file(s).
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: pcm2wav.c,v 1.2 2004/04/20 22:43:37 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string>
#include <cerrno>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
#include <AH/IO/Riff/WaveFile.h>
USING_NAMESPACE(AH)

// --- local constants ---

static const char* cmdLineOptions = "c:f:hi:r:s:vV";
static const char* version = "1.0";
static const uint32 ONE_MB = 1 * 1024 * 1024;  // 1MB
static const uint32 MAX_BUF_SIZE = ONE_MB;  // 1MB

// --- local classes ---

class AudioFormat
{
 public:
    AudioFormat()
	: channelNum(2), sampleSize(16), sampleRate(44100), 
          fileSize(1024 * ONE_MB) {}

    uint32 getSampleNum() const
    {
	uint32 dataSize = fileSize - 0x200;
	uint32 chanSize = (channelNum * sampleSize) / 8;
	return dataSize / chanSize;
    }

    uint16 channelNum;	// number of channels
    uint16 sampleSize;	// size of sample in bits
    uint32 sampleRate;	// sample rate in Hz
    uint32 fileSize;    // size of file in bytes
};	

class DataCopyRaw
{
 public:
    /// constructor
    DataCopyRaw(FILE* inputStream, const string& inputName,
		IORiffWaveFile& outputStream, const string& outputName,
		int v = 0)
	: verbose(v), in(inputStream), inputFileName(inputName),
	  out(outputStream), outputFileName(outputName) {}

    /// copy specified data range
    bool copy(uint32 size);

 protected:

    /// read data buffer
    int readBuf(unsigned char* buf, unsigned int size);

    /// write data buffer
    int writeBuf(const unsigned char* buf, unsigned int size);

    int verbose;

 private:
    FILE* in;
    const string& inputFileName;
    IORiffWaveFile& out;
    const string& outputFileName;

}; // class DataCopyRaw


// read data buffer
int DataCopyRaw::readBuf(unsigned char* buf, unsigned int size)
{
    if (4 < verbose) cerr << "read " << size 
			  << " bytes from input file \"" 
			  << inputFileName << "\":" << endl;
    int32 rdsize = fread(buf, 1, size, in);
    return rdsize;
}

// write data buffer
int DataCopyRaw::writeBuf(const unsigned char* buf, unsigned int size)
{
    uint32 sizeWritten = 0;
    uint32 sizeToWrite = size;
    while (0 < sizeToWrite)
    {
	if (4 < verbose) cerr << "write " << sizeToWrite 
			      << " bytes into output file \"" 
			      << outputFileName << "\":" << endl;
	sizeWritten = out.writeRaw(buf, sizeToWrite);
	if (0 >= sizeWritten) break;
	sizeToWrite -= sizeWritten;
    }
    if (0 == sizeToWrite) return size;
    return sizeWritten;
}


//-----------------------------------------------------------------------
// copy specified data from input file to output file (based on samples)
// On Error: throws an exception
//-----------------------------------------------------------------------
bool DataCopyRaw::copy
(
    uint32 size   // size to copy
)
{
    if (verbose > 4)
	cerr << "Copy " << size << " bytes" << endl;

    // make a buffer for R/W of data (on stack):
    const uint32 bufSize = MAX_BUF_SIZE;
    unsigned char buf[bufSize];

    uint32 sizeToRead = size;
    do
    {
	// how much can we read (not more than buffer size) ?
	uint32 curBufSize = sizeToRead;
	if (curBufSize > bufSize) curBufSize = bufSize;

	// read input file data :
	int32 rdsize = readBuf(buf, curBufSize);
	if (0 >= rdsize) return false;
	sizeToRead -= rdsize;
	
	// write output file data :
	if (0 >= writeBuf(buf, rdsize)) 
	    throw ErrorBasic("Cannot write WAV file");
    } while (sizeToRead);
    return true;
} // end of copy_data()





// --- local variables ---

static AudioFormat fmt;
static int verbose = 0;

// --- local functions ---


static void usage(bool help = false)
{
    cerr << "Usage: pcm2wav [-" << cmdLineOptions << "] file\n"
	 << "       cut a piece from WAV file(s) into new WAV file.\n"
	 << "  options:\n"
	 << "   -h                  - print this help message\n"
	 << "   -f <filesize>       - specify size of WAV file in MB\n"
	 << "                         (default is " << fmt.fileSize/ONE_MB << ")\n"
	 << "   -i <input file>     - specify input file\n"
	 << "   -c <channels>       - specify number of channels\n"
	 << "                         (default is " << fmt.channelNum << ")\n"
	 << "   -r <samplerate>     - specify samplerate in kHz\n"
	 << "                         (default is " << (double)fmt.sampleRate/1000 << ")\n"
	 << "   -s <samplesize>     - specify samplesize in bits\n"
	 << "                         (default is " << fmt.sampleSize << ")\n"
	 << "   -v                  - verbose mode (increase debug level)\n"
	 << "   -V                  - print version" << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << "  pcm2wav -i data.pcm data\n"
	 << "  - convert file data.pcm file(s) data-1.wav, data-2.wav, ...\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "pcm2wav Vs " << version << " " 
	 << "(c) 2000 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//----------------------------------------------------------------
// copy specified parts of all input files into output file
// On Error: throws an exception
//----------------------------------------------------------------
static void process_files
(
    const AudioFormat& fmt,
    const string& inputFileName,
    const string& outputFileName
)
{
    if (2 < verbose) cerr << "process files" << endl;
    if (3 < verbose) cerr << "create output file \"" << outputFileName
			  << "\":" << endl;
    FILE* fp = stdin;
    if ("-" != inputFileName)
    {
	fp = fopen(inputFileName.c_str(),"r");
	if (!fp) throw ErrorSys(inputFileName);
    }

    try 
    {
	int fileNum = 1;
	bool finished;
	do
	{
	    // determine name of output file:
	    char ext[64];
	    (void)sprintf(ext,"-%02d.wav", fileNum);
	    string curOutName = outputFileName + ext;

	    IORiffWaveFile out(curOutName.c_str(), fmt.channelNum,
			       fmt.sampleSize, fmt.sampleRate,
			       fmt.getSampleNum());
            uint32 sizeToRead = fmt.fileSize - 0x200;

	    DataCopyRaw data(fp, inputFileName, out, curOutName, verbose);
	    finished = !data.copy(sizeToRead);

	    out.writeHeader();
	    ++fileNum;
	} while(!finished);
    }
    catch (...)
    {
	if ("-" != inputFileName) (void)fclose(fp);
	throw;
    }
} // end of process_files()


// --- global functions ---


int main(int argc, char* argv[])
{
    //------------------------------
    // handle command line options:
    //------------------------------
    string inputFileName("-");
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'c': fmt.channelNum = atoi(optarg);                  break;
	case 'f': fmt.fileSize = ONE_MB * atoi(optarg);           break;
	case 'h': usage(true);                                    exit(0);
	case 'i': inputFileName = optarg;                         break;
	case 'r': fmt.sampleRate = (uint32)(1000 * atof(optarg)); break;
	case 's': fmt.sampleSize = atoi(optarg);                  break;
	case 'v': verbose++;                                      break;
	case 'V': print_version();                                exit(0);
	default : usage();                                        exit(1);
	}
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing output file" << endl;
	usage();
	exit(1);
    }
    // the last argument is the output file name;
    string outputFileName(argv[optind++]);
    // check for too many arguments:
    if (optind < argc) 
    {
	cerr << "Error: too many arguments (last valid is \"" 
	     << outputFileName << "\")" << endl;
	usage();
	exit (1);
    }

    //--------------------------
    // process all input files:
    //--------------------------
    try { process_files(fmt, inputFileName, outputFileName); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
