//---------------------------------------------------------------------------
// Copyright 2005 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/wavtool/wavbitred.c,v $
// $Id: wavbitred.c,v 1.2 2005/02/11 18:51:30 ah Exp $ 
//
// Description: reduce number of bits (w/wo dither) of a wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: wavbitred.c,v 1.2 2005/02/11 18:51:30 ah Exp $ (c) 2005 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string>
#include <cerrno>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/IO/Convert.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/Math/Dither.h>
USING_NAMESPACE(AH)

// --- local constants ---

static const char* cmdLineOptions = "b:d:hvV";
static const char* version = "0.2";

// reduced as cygwin has problems with big arrays on the stack
// static const uint32 MAX_SAMPLES = 256 * 1024;  // 256k
static const uint32 MAX_SAMPLES = 64 * 1024;  // 64k

// --- local variables ---

static int verbose = 0;

// --- local classes ---

// --- local functions ---

static void usage(bool help = false)
{
    cerr << "Usage: wavbitred [-" << cmdLineOptions << "] infile outfile\n"
	 << "       reduce bit size in a WAV file.\n"
	 << "  options:\n"
	 << "   -h       - print this help message\n"
	 << "   -b <num> - number of remaining bits (default = 16)\n"
	 << "    d <typ> - add dither\n"
	 << "              typ=r: rectangulare PDF\n"
	 << "              typ=t: triangle PDF\n"
	 << "              typ=n: no dither, round to next Bit (default)\n"
	 << "   -v       - verbose mode (increase debug level)\n"
	 << "   -V       - print version\n" 
	 << "  infile    - WAV file for input\n" 
	 << "  outfile   - WAV file for output\n" 
	 << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "wavbitred Vs " << version << " " 
	 << "(c) 2005 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static void process
(
    const string& inputFileName,
    const string& outputFileName,
    int bits,
    MathDither::Type d
)
{
    if (2 < verbose) cerr << "process:" << endl;

    // create a wav file object for read:
    if (3 < verbose) cerr << "open input file \"" << inputFileName
			  << "\":" << endl;
    IORiffWaveFile in(inputFileName.c_str(), false, verbose);
    if (verbose  > 3) in.printInfo(cerr);

    // create a wav file object for write:
    if (3 < verbose) cerr << "create output file \"" << outputFileName
			  << "\":" << endl;
    IORiffWaveFile out(outputFileName.c_str(), in.getChannelNum(),
		       in.getSampleSize(), in.getSampleRate(),
		       in.getSampleNum());

    // create the Dither object:
    MathDither dith(d);

    int bitred = in.getSampleSize() - bits;
    if (bitred < 0) throw ErrorBasic("illegal bit size");
    int redfact = 0x1 << bitred;

    // get size of data buffer for desired samples
    const uint32 bufSize = in.getDataBufSize(MAX_SAMPLES);
    unsigned char buf[bufSize];

    // read without manipulation
    uint32 samplesToRead = in.getSampleNum();
    uint32 sizeToRead = in.getDataBufSize(samplesToRead);
    do
    {
	// how much can we read (not more than buffer size) ?
	uint32 curSamplesToRead = samplesToRead;
	if (curSamplesToRead > MAX_SAMPLES) curSamplesToRead = MAX_SAMPLES;
	uint32 curBufSize = in.getDataBufSize(curSamplesToRead);

	uint32 samNumToWrite = 0;
	uint32 sizeToWrite = 0;

	// read input file data :
	int32 rdsize = in.readRaw(buf, curBufSize);
	if (0 >= rdsize) break;
	sizeToRead -= rdsize;

	// process data:
	// get specified sample from buffer
	uint32 samNum = MAX_SAMPLES;
	if ((uint32)rdsize < bufSize)
	    samNum = rdsize / in.getDataBufSize((uint32)1);
	samNumToWrite = samNum; 
	sizeToWrite = rdsize; 

	for (uint32 i = 0; i < samNum; ++i)
	{
	    for (uint32 chan = 0; chan < in.getChannelNum(); ++chan)
	    {
		int32 value = in.getBufSample(buf, i, chan); 
		double dval = dith.getSignal();
		double redval = (double)value / (double)redfact + dval;
		double rnd = 0.5;
		if (value < 0) rnd = -0.5;
		value = (int32)(redval + rnd);
		value *= redfact; 
		out.setBufSample(buf,i, chan, value);
	    }
	}

	// write the buffer into the output file:
	uint32 sizeWritten = 0;
	while (0 < sizeToWrite)
	{
	    if (4 < verbose) cerr << "write " << sizeToWrite 
				  << " bytes into output file \"" 
				  << outputFileName << "\":" << endl;
	    sizeWritten = out.writeRaw(buf, sizeToWrite);
	    if (0 >= sizeWritten) break;
	    sizeToWrite -= sizeWritten;
	}
	if (sizeToWrite)
	    cerr << "Warning: could not write " << sizeToWrite  
		 << " bytes into output file \"" 
		 << outputFileName << "\"" << endl;
    }
    while (sizeToRead > 0);
    out.writeHeader();
} // end of process()


// --- global functions ---


int main(int argc, char* argv[])
{
    int bits = 16;
    MathDither::Type dither = MathDither::None;

    //------------------------------
    // handle command line options:
    //------------------------------
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'b': bits = atoi(optarg); break;
	case 'd': switch(optarg[0])
	          {
		  default:
		  case 'n': dither = MathDither::None; break;
		  case 'r': dither = MathDither::RectAngle; break;
		  case 't': dither = MathDither::Triangle; break;
		  }
	          break;
	case 'v': verbose++;            break;
	case 'h': usage(true);          exit(0);
	case 'V': print_version();      exit(0);
	default : usage();              exit(1);
	}
    }

    // handle the remaining 2 arguments: 
    if (argc < optind + 2) 
    {
	cerr << "Error: missing arguments" << endl;
	usage();
	exit(1);
    }
    string inputFileName(argv[optind++]);
    string outputFileName(argv[optind++]);

    // check for bad arguments:
    if (bits < 1)
    {
	cerr << "Error: argument to option -b is out of range (>0)" << endl;
	usage();
	exit (1);
    }

    // check for too many arguments:
    if (optind < argc) 
    {
	cerr << "Error: too many arguments (last valid is \"" 
	     << argv[optind -1] << "\")" << endl;
	usage();
	exit (1);
    }

    //---------------------
    // process input file:
    //---------------------
    try 
    {
	process(inputFileName, outputFileName, bits, dither); 
    }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
