//---------------------------------------------------------------------------
// Copyright 2004 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/wavtool/wavspec.c,v $
// $Id: wavspec.c,v 1.19 2005/02/17 22:55:18 ah Exp $ 
//
// Description: create frequency spectrum of a wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: wavspec.c,v 1.19 2005/02/17 22:55:18 ah Exp $ (c) 2004 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string>
#include <cerrno>
#include <cmath>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <drfftw.h> 
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
#include <AH/STL/String.h>
#include <AH/IO/Convert.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/IO/Riff/WaveFileCached.h>
#include <AH/IO/Plot.h>
#include <AH/IO/CmdOption.h>
#include <AH/IO/CmdOptHandler.h>
#include <AH/IO/TimeStamp.h>
#include <AH/IO/Smart/Buffer.h>
#include <AH/Math/Ft/Wisdom.h>
#include <AH/Math/DataStat.h>
#include <AH/Math/ReduceSpectrum.h>
USING_NAMESPACE(AH)

// --- local definitions ---

#define PROFILING 0
#define TRACE 0
#define ACC 1

// --- local constants ---

static const char* version = "0.8";
static uint32 MAX_SAMPLES = 1 << (16);  // 64k
static uint32 OUT_SAMPLES = 1 << (10);  // 1024

// --- local variables ---

static int verbose = 0;
#if PROFILING
static int profiling = 0;
#endif

// --- local classes ---

// --- local functions ---

static void usage(IOCmdOptHandler& opt, bool help = false)
{
    cerr << "Usage: wavspec " << opt.getShortOptionText() 
	 << " <file> [<file>, ..]\n"
	 << opt.getHelpText("  ")
	 << "   <file>         - audio file, supported file types:\n"
	 << "                     .wav           : PCM-WAV\n"
	 << "                     .mp3           : MP3\n"
	 << "                     .mpc .mp+ .mpp : MPEG+\n"
	 << "                     .ogg           : OGG-Vorbis\n"
	 << "                     .pac           : LPAC"
	 << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << "  Create mean spectrum of file aaa.wav:\n"
	 << "    wavspec aaa.wav\n"
	 << "  Create mean spectrum of file aaa.wav and watch its change:\n"
	 << "    wavspec -c aaa.wav\n"
	 << "  Create mean spectrum of file aaa.wav smoothed over 1 octave:\n"
	 << "    wavspec -s 1 aaa.wav\n"
	 << "  Create peak spectrum of file aaa.wav:\n"
	 << "    wavspec -p aaa.wav\n"
	 << "\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "wavspec Vs " << version << "\n" 
	 << "  (c) 2004 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to Andreas Hnnebeck <ah@despammed.com>.\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//------------------------------------------------------------------
// for debugging: calculate min, max and mean value from data array
//------------------------------------------------------------------
static void print_min_max_mean(const std::string& head, 
			       double* data, unsigned int size)
{
    // determine min, mean and max value:
    double mean;
    double max;
    double min;
    DataStat::calcMeanMaxMin(data, size, &mean, &max, &min);
    cerr << head << ": min = " << min << ", mean = " << mean 
	 << ", max = " << max << endl;
}

// make fourier transformation
static void make_ft(rfftw_plan& plan, fftw_real* data, unsigned int& size)
{
    if (verbose > 1) cerr << "executing FFT" << endl;
    if (verbose > 2) print_min_max_mean("time domain", data, size);

    //----------------------------------------
    // calculate the FFT from the input data:
    //----------------------------------------
    IOSmartBuffer<fftw_real> dout(size);
    rfftw_one(plan, data, dout());
    unsigned int snum = size / 2;

    // put result back into data array:
    // - put back absolute values
    data[0] = fabs(dout[0]) / snum;
    for (unsigned int i = 1; i < snum; ++i)
    {
	data[i] = sqrt(dout[i]*dout[i] + dout[size-i]*dout[size-i]) / snum;
    }
    data[snum] = fabs(dout[snum] / snum);
    size = snum;

    if (verbose > 2) print_min_max_mean("frequency domain", data, size);
} // make_ft()


// convert amplitude into dB
void make_ylog_scale(double* data, unsigned int size,
		     double maxamp, double minval, double mindb)
{
    if (verbose > 1) cerr << "convert amplitude into dB" << endl;

    // convert absolute values into dB
    for (unsigned int i = 0; i < size; ++i)
    {
	double val = data[i];
	data[i] = (val < minval ? mindb : 20 * log10(val/maxamp));
	if (verbose > 3) cerr << "lin = " << val 
			      << " log = " << data[i] << endl;
    }
} // make_ylog_scale()



// make the plot
static std::string get_data_plot_cmd
(
    double* mean[],
    double* peak[],
    unsigned int* dataSize,
    unsigned int ch_num,
    double f1, 
    double f2,
    bool is_log
)
{
    // plot data in buffer:
    // - we cannot use sstream, this is approximately 150 times
    //   slower than this mix of STL-string and sprintf
    std::string cmd("plot ");
    for (unsigned int chan = 0; chan < ch_num; ++chan)
    {
	if (chan) cmd += ", ";
	if (mean && peak)
	{
	    cmd += STLStringPrintf(80, "\"-\" using 1:2 title \"ch%d peak\", ",
				   chan+1);
	    cmd += STLStringPrintf(80, "\"-\" using 1:2 title \"ch%d mean\"",
				   chan+1);
	}
	else
	    cmd += STLStringPrintf(80, "\"-\" using 1:2 title \"ch%d\"", 
				   chan+1);
    }
    cmd += "\n";
    if (is_log)
    {
	if (peak) 
	{
	    for (unsigned int chan = 0; chan < ch_num; ++chan)
	    {
		double f = f1;
		const double f_inc = pow(f2/f1, 1.0 / (dataSize[chan] - 1.0));
		for (unsigned int i = 0; i < dataSize[chan]; ++i, f *= f_inc)
		{
		    cmd += STLStringPrintf(80,"%f %f\n", f, peak[chan][i]);
		}
		cmd += "e\n";
	    }
	}
	if (mean) 
	{
	    for (unsigned int chan = 0; chan < ch_num; ++chan)
	    {
		double f = f1;
		const double f_inc = pow(f2/f1, 1.0 / (dataSize[chan] - 1.0));
		for (unsigned int i = 0; i < dataSize[chan]; ++i, f *= f_inc)
		{
		    cmd += STLStringPrintf(80,"%f %f\n", f, mean[chan][i]);
		}
		cmd += "e\n";
	    }
	}
    }
    else
    {
	if (peak)
	{
	    for (unsigned int chan = 0; chan < ch_num; ++chan)
	    {
		double f = f1;
		const double f_inc = (f2 - f1) / (dataSize[chan] - 1.0);
		for (unsigned int i = 0; i < dataSize[chan]; ++i, f += f_inc)
		{
		    cmd += STLStringPrintf(80,"%f %f\n", f, peak[chan][i]);
		}
		cmd += "e\n";
	    }
	}
	if (mean)
	{
	    for (unsigned int chan = 0; chan < ch_num; ++chan)
	    {
		double f = f1;
		const double f_inc = (f2 - f1) / (dataSize[chan] - 1.0);
		for (unsigned int i = 0; i < dataSize[chan]; ++i, f += f_inc)
		{
		    cmd += STLStringPrintf(80,"%f %f\n", f, mean[chan][i]);
		}
		cmd += "e\n";
	    }
	}
    }
    return cmd;
} // get_data_plot_cmd



//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static void process
(
    const std::string& inputFileName,
    const std::string& plotfile,
    const std::string& title,
    const std::vector<std::string>& settings,
    bool mean,
    bool peak,
    bool continuous,
    bool is_log,
    double smooth
)
{
#if PROFILING
    IOTimeStamp startTime;
    IOTimeStamp lastTime;
    IOTimeStamp curTime;
#endif
    if (2 < verbose) 
	cerr << "process(\"" << inputFileName << ", \"" << plotfile 
	     << ", \"" << title << ",, -c=" << continuous 
	     << ", freqscale=:" << is_log << ", -s=" << smooth << ")" 
	     << endl;

    //--------------
    // read values:
    //--------------
    IORiffWaveFileCached in(inputFileName.c_str(), MAX_SAMPLES, verbose);
    if (verbose  > 3) in.printInfo(cerr);
    const unsigned int ch_num = in.getChannelNum();
    const double maxamp = 1 << (in.getSampleSize() -1);
    const double minval = 1.0 / 64; // 6 Bits below LSB (due to dither)
    const double mindb = 20 * log10(minval/maxamp);
    const double f_max = in.getSampleRate() / 2;
    const double dwelltime = 1.0 / in.getSampleRate();
    const double buftime = dwelltime * MAX_SAMPLES;

    // create input data array for each channel:
    // - and corresponding size array
    IOSmartBuffer<unsigned int> dataSize(ch_num);
    fftw_real** data = new fftw_real*[ch_num];
    if (!data) throw ErrorSys(ENOMEM);
    for (unsigned int chan = 0; chan < ch_num; ++chan)
    {
	dataSize[chan] = MAX_SAMPLES;
	data[chan] = new fftw_real[dataSize[chan]];
	if (!data[chan]) throw ErrorSys(ENOMEM);
    }
    double f1 = 20;
    double f2 = 20000;

    // create plot object:
    IOPlot plot;

    // set the default settings:
    std::vector<std::string> plotcmd;
    plotcmd.push_back("set grid");
    plotcmd.push_back("set tics out");
    plotcmd.push_back("set key bottom right");
    if (OUT_SAMPLES < 70)
	plotcmd.push_back("set data style steps");
    else
	plotcmd.push_back("set data style lines");
    plotcmd.push_back("set xlabel \"Hz\"");
    plotcmd.push_back("set ylabel \"dB\"");
    plotcmd.push_back("set title \"" + inputFileName + ": " + title + "\"");
    if (is_log)
    {
	plotcmd.push_back("set logscale x");
	plotcmd.push_back("set xtics ("
			  "\"1\" 1, \"2\" 2, \"5\" 5, "
			  "\"10\" 10, \"20\" 20, \"50\" 50, "
			  "\"100\" 100, \"200\" 200, \"500\" 500, "
			  "\"1k\" 1000, \"2k\" 2000, \"5k\" 5000, "
			  "\"10k\" 10000, \"20k\" 20000, \"50k\" 50000)");
    }
    char str[80];
    sprintf(str, "set xrange [%d:%d]", (int)f1, (int)f2);
    plotcmd.push_back(str);
    sprintf(str, "set yrange [%.0f:20]", mindb-10);
    plotcmd.push_back(str);
    plot.exec(plotcmd);

    // now add the additional settings
    plot.exec(settings);

    // get ft wisdom:
    MathFtWisdom wis(verbose);
    rfftw_plan plan = wis.getPlan(MAX_SAMPLES);

    //----------------------
    // read the whole file:
    //----------------------
    if (in.getSampleSize() == 16)
    {
	//--------------------------------------------------------
	// use fast read method for files with 16 Bit Samplerate:
	//--------------------------------------------------------

	// create buffer for accumulated mean data:
	double** accum = new double*[ch_num];
	if (!accum) throw ErrorSys(ENOMEM);
	for (unsigned int chan = 0; chan < ch_num; ++chan)
	{
	    accum[chan] = new double[OUT_SAMPLES];
	    if (!accum[chan]) throw ErrorSys(ENOMEM);
	    for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
		accum[chan][i] = 0;
	}

	double** accum_peak = 0;
	if (peak)
	{
	    // create buffer for accumulated peak data:
	    accum_peak = new double*[ch_num];
	    if (!accum_peak) throw ErrorSys(ENOMEM);
	    for (unsigned int chan = 0; chan < ch_num; ++chan)
	    {
		accum_peak[chan] = new double[OUT_SAMPLES];
		if (!accum_peak[chan]) throw ErrorSys(ENOMEM);
		for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
		    accum_peak[chan][i] = 0;
	    }
	}

	const uint32 bufSize = MAX_SAMPLES * in.getChannelNum() * 2;
	int16 buf[bufSize];
	double runtime = 0;
	unsigned int cnt = 0;
	ReduceSpectrum reduce(verbose, OUT_SAMPLES, f_max, f1, f2);
	while(1)
	{
	    // read buffer:
	    uint32 curBufRead = in.readRaw((unsigned char*)buf, bufSize);
	    if (curBufRead < bufSize) 
		break; // ignore last piece
	    ++cnt;

	    int16* value = buf;
	    for (uint32 i = 0; i < MAX_SAMPLES; ++i)
	    {
		for (unsigned int chan = 0; chan < ch_num; ++chan)
		{
		    data[chan][i] = *value++;
		}
	    }
#if PROFILING
	    if (profiling > 1) lastTime.update();
#endif
	    for (unsigned int chan = 0; chan < ch_num; ++chan)
	    {
		dataSize[chan] = MAX_SAMPLES; // restore data size
		make_ft(plan, data[chan], dataSize[chan]);

		// store spectrum in buffer:
		double spec_buf[MAX_SAMPLES]; // spectrum buffer
		for (unsigned int i = 0; i < dataSize[chan]; ++i)
		    spec_buf[i] = data[chan][i];

		double mean_buf[OUT_SAMPLES]; // mean buffer
		double peak_buf[OUT_SAMPLES]; // peak buffer
		
		// make peak and mean data:
		if (is_log)
		{
		    reduce.processLogXScale(data[chan], dataSize[chan],
					    mean_buf, peak_buf, smooth);
		}
		else
		{
		    reduce.processLinXScale(data[chan], dataSize[chan],
					    mean_buf, peak_buf);
		}
		dataSize[chan] = OUT_SAMPLES;

		if (mean)
		{
		    // keep mean data:
		    for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
			accum[chan][i] += mean_buf[i];
		}

		if (peak)
		{
		    // keep peak data:
		    for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
			accum_peak[chan][i] += peak_buf[i];
		}

		if (continuous)
		{
		    if (peak)
		    {
			// display peak data:
			for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
#if ACC
			    data[chan][i] = accum_peak[chan][i] / cnt;
#else
			    data[chan][i] = peak_buf[i];
#endif
		    }
		    else
		    {
			// display mean data:
			for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
#if ACC
			    data[chan][i] = accum[chan][i] / cnt;
#else
			    data[chan][i] = mean_buf[i];
#endif
		    }
		    make_ylog_scale(data[chan], dataSize[chan],
				    maxamp, minval, mindb);
		}
	    }
#if PROFILING
	    if (profiling > 1)
	    {
		curTime.update();
		cerr << "time for ft: " 
		     << (curTime - lastTime).getTimeString() << endl;
		lastTime.update();
	    }
#endif
	    if (continuous)
	    {
		// plot data in buffer:
		// - we cannot use sstream, this is approximately 150 times
		//   slower than this mix of STL-string and sprintf
		std::string cmd = "set title \"" + title;
		char cbuf[80];
		sprintf(cbuf, " %.1fs\"\n", runtime);
		cmd += cbuf;
		cmd += get_data_plot_cmd(data, 0, dataSize(), ch_num, f1, f2,
					 is_log);
#if PROFILING
		if (profiling > 1)
		{
		    curTime.update();
		    cerr << "time for command: " 
			 << (curTime - lastTime).getTimeString() << endl;
		    lastTime.update();
		}
#endif
		plot.exec(cmd);
#if PROFILING
		if (profiling > 1)
		{
		    curTime.update();
		    cerr << "time for plot: " 
			 << (curTime - lastTime).getTimeString() << endl;
		}
#endif
	    }
	    runtime += buftime;
	}
	if (0 == cnt)
	{
	    cerr << "Warning: not enough data available in \"" 
		 << inputFileName << "\". Decreasing samplenum may help"
		 << endl;
	    return;
	}

	// process accumulated data for average spectrum:
	for (unsigned int chan = 0; chan < ch_num; ++chan)
	{
	    if (mean)
	    {
		for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
		    accum[chan][i] /= cnt;
		make_ylog_scale(accum[chan], OUT_SAMPLES, maxamp, 
				minval, mindb);
	    }
	    if (peak)
	    {
		for (unsigned int i = 0; i < OUT_SAMPLES; ++i)
		    accum_peak[chan][i] /= cnt;
		make_ylog_scale(accum_peak[chan], OUT_SAMPLES, maxamp, 
				minval, mindb);
	    }
	}

	// plot accumulated data in buffer:
	// - we cannot use sstream, this is approximately 150 times
	//   slower than this mix of STL-string and sprintf
	std::string cmd;
	cmd = "set title \"" + inputFileName + ": " + title;
	if (peak) cmd += " (peak averaged over ";
	else      cmd += " (mean average over ";
	cmd += STLStringPrintf(80, "%.1fs)\"\n", runtime);
	plot.exec(cmd);
	cmd = get_data_plot_cmd(mean ? accum : 0, peak ? accum_peak : 0,
				dataSize(), ch_num, f1, f2, is_log);
	plot.exec(cmd);

	if (plotfile != "")
	{
	    // determine extension of plotfile:
	    string::size_type pos = plotfile.find_last_of(".");
	    if (string::npos != pos)
	    {
		// extension found:
		std::string terminal = plotfile.substr(pos+1);

		if (terminal == "png")
		{
		    terminal += " color";
		    plot.exec("set size 1024./128./5., 768./120./4.");
		}
		else if (terminal == "ps")
		{
		    terminal = "postscript monochrome";
		    plot.exec("set size 1,1");
		}
		else if (terminal == "eps")
		{
		    terminal = "postscript eps color";
		    plot.exec("set size 1,1");
		}

		plot.exec("set terminal " + terminal);
		if (is_log)
		{
		    plot.exec("set xtics ("
			"\"1\" 1, \"1.5\" 1.5, \"2\" 2, \"3\" 3, " 
			      "\"4\" 4, \"5\" 5, \"7\" 7, "
			"\"10\" 10, \"15\" 15, \"20\" 20, \"30\" 30, "
			      "\"40\" 40, \"50\" 50, \"70\" 70, "
			"\"100\" 100, \"150\" 150, \"200\" 200, \"300\" 300, "
			      "\"400\" 400, \"500\" 500, \"700\" 700, "
			"\"1k\" 1000, \"1k5\" 1500, \"2k\" 2000, \"3k\" 3000, "
			      "\"4k\" 4000, \"5k\" 5000, \"7k\" 7000, "
			"\"10k\" 10000, \"15k\" 15000, \"20k\" 20000, "
			      "\"30k\" 30000, \"40k\" 40000, \"50k\" 50000)");
		}
		plot.exec("set output \"" + plotfile);
		plot.exec(cmd);
		if (verbose)
		    cerr << "created \"" << plotfile << "\"" << endl;
	    }
	}
	for (unsigned int chan = 0; chan < ch_num; ++chan)
	    delete[] accum[chan];
	delete[] accum;
    }
    else
    {
#if 1
	cerr << "Other Samplesizes than 16 Bit are not yet implemented."
	     << endl;
#else
	//---------------------------------------------------------------
	// use standard read method for files with difficult Samplerate:
	//---------------------------------------------------------------

	// how many samples do we have to read:
	uint32 samToRead = in.getSampleNum();

	for (uint32 curSam = 0; curSam < samToRead; ++curSam)
	{
	    for (uint32 chan = 0; chan < in.getChannelNum(); ++chan)
	    {
		// read value:
		int32 val;
		if (0 >= in.readSample(&val, curSam, chan))
		    break;

		// increment array field of matching value
		++signedValue[val];
		++total;
		if (val < minVal) minVal = val; 
		if (val > maxVal) maxVal = val; 
	    }
	}
#endif
    }

    for (unsigned int chan = 0; chan < ch_num; ++chan)
	delete[] data[chan];
    delete[] data;

#if PROFILING
    if (profiling)
    {
	curTime.update();
	cerr << "total run time: " 
	     << (curTime - startTime).getTimeString() << endl;
    }
#endif
    cerr << "hit return to continue" << endl;
    char tmp[80];
    (void)fgets(tmp, 79, stdin);
} // end of process()


// --- global functions ---


int main(int argc, char* argv[])
{
    // set all options:
    IOCmdOption OptionVerbose('v', "verbose", IOCmdOption::SecondArgNone,
	"increase verbose level (default=off)");
    IOCmdOption OptionVersion('V', "version", IOCmdOption::SecondArgNone,
	"print version and exit");
    IOCmdOption OptionHelp('h', "help", IOCmdOption::SecondArgNone,
	"print this help message");
    IOCmdOption OptionContinuous('c', "cont", IOCmdOption::SecondArgNone,
	"do continous plots while reading file");
    IOCmdOption OptionMean('m', "mean", IOCmdOption::SecondArgNone,
	"calculate mean spectrum (default)");
    IOCmdOption OptionPeak('p', "peak", IOCmdOption::SecondArgNone,
	"calculate peak spectrum");
    IOCmdOption OptionGnuplot('g', "gnuplot", IOCmdOption::SecondArgAlways,
	"specify setting for gnuplot (may be given multiply)");
    IOCmdOption OptionOutput('o', "output", IOCmdOption::SecondArgAlways,
	"specify plot output file");
    IOCmdOption OptionOutSize('\0', "pointnum", IOCmdOption::SecondArgAlways,
	"number of points in plot (default=1024)");
    IOCmdOption OptionInSize('\0', "samplenum", IOCmdOption::SecondArgAlways,
	"number of samples in kB to read in one block (default=64k)");
    IOCmdOption OptionFreqScale(
	'\0', "freqscale", IOCmdOption::SecondArgAlways,
	"set frequency scale (lin or log, default=log)");
    IOCmdOption OptionSmooth('s', "smooth", IOCmdOption::SecondArgAlways,
	"smooth plot over specified range in octaves");
    IOCmdOption OptionPlotTitle('t', "title", IOCmdOption::SecondArgAlways,
	"title of plot");
#if PROFILING
    IOCmdOption OptionProfile('p', "profile", IOCmdOption::SecondArgNone,
	"increase profiling level (default=none)");
#endif

    // load all options into option handler:
    IOCmdOptHandler hdl;
    hdl.addOption(&OptionHelp);
    hdl.addOption(&OptionContinuous);
    hdl.addOption(&OptionMean);
    hdl.addOption(&OptionPeak);
    hdl.addOption(&OptionGnuplot);
    hdl.addOption(&OptionOutput);
    hdl.addOption(&OptionSmooth);
    hdl.addOption(&OptionPlotTitle);
#if PROFILING
    hdl.addOption(&OptionProfile);
#endif
    hdl.addOption(&OptionVerbose);
    hdl.addOption(&OptionVersion);
    hdl.addOption(&OptionInSize);
    hdl.addOption(&OptionOutSize);
    hdl.addOption(&OptionFreqScale);

    // get all option settings from command line arguments:
    // - the remaining arguments are input files
    std::vector<std::string> in_fn;
    std::vector<std::string> errwarn;
    if (0 != hdl.handleArgs(argc, argv, in_fn, errwarn))
    {
	if (errwarn.size())
	{
	    for (unsigned int i = 0; i < errwarn.size(); ++i)
		cerr << "  Attention: " << errwarn[i] << endl;
	}
	usage(hdl, false);
	exit(1); // abort on bad call
    }

    // handle all options:

    if (OptionHelp.isSet())
    {
	usage(hdl, true);
	exit(0);
    }

    if (OptionVersion.isSet())
    {
	print_version();
	exit(0);
    }

    verbose = OptionVerbose.isSet();
#if PROFILING
    profiling = OptionProfile.isSet();
#endif

    std::string title("Frequency Spectrum");
    if (OptionPlotTitle.isSet())
    {
	std::string t = OptionPlotTitle.getLastSecArg();
	if (t != "") title = t;
    }

    if (OptionInSize.isSet())
    {
	unsigned int s = atoi(OptionInSize.getLastSecArg().c_str());
	if (s) MAX_SAMPLES = s * 1024;
    }
    if (OptionOutSize.isSet())
    {
	unsigned int s = atoi(OptionOutSize.getLastSecArg().c_str());
	if (s) OUT_SAMPLES = s;
    }

    bool is_log = true;
    if (OptionFreqScale.isSet())
    {
	std::string t = OptionFreqScale.getLastSecArg();
	if (t == "lin") is_log = false;
	else if (t != "log")
	{
	    cerr << "Error: bad argument \"" << t << "\" for option --"
		 << OptionFreqScale.getLongOpt() << endl;
	    usage(hdl, false);
	    exit(1);
	}
    }

    double smooth = 0;
    if (OptionSmooth.isSet())
    {
	smooth = atof(OptionSmooth.getLastSecArg().c_str());
    }

    bool mean = OptionMean.isSet();
    bool peak = OptionPeak.isSet();
    if (!mean && !peak) mean = true; // set default

    // check input files:
    if (0 == in_fn.size()) 
    {
	cerr << "Error: missing input file" << endl;
	usage(hdl, false);
	exit(1);
    }

    //---------------------
    // process input file:
    //---------------------
    try 
    { 
	// process all files
	for (unsigned int i = 0; i < in_fn.size(); ++i)
	{
	    try
	    {
		// process files
		process(in_fn[i], OptionOutput.getLastSecArg(), title,
			OptionGnuplot.getSecArg(),
			mean, peak,
			0 < OptionContinuous.isSet(),
			is_log, smooth);
	    }
	    catch(const ErrorClass& err)
 	    {
		cerr << "Error handling '" << in_fn[i] << "' :" 
		     << err.getMsg() << endl;
	    }
	    catch(const exception& err)
	    {
		cerr << "Error handling '" << in_fn[i] << "' :" 
		     << err.what() << endl;
	    }
	    catch(...)
	    {
		cerr << "Error handling '" << in_fn[i] 
		     << "' : unknown exception" << endl;
	    }
	}
    }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
